@@ -6,13 +6,28 @@ In basically all cases, you do not need them.
 from __future__ import annotations
 
 import copy
+import dataclasses
 import inspect
 import sys
 import warnings
 from collections import defaultdict
 from functools import wraps
 from types import MethodWrapperType
-from typing import Any, Callable, DefaultDict, Dict, Generic, Iterable, List, Optional, Tuple, Type, TypeVar, Union
+from typing import (
+    Any,
+    Callable,
+    ClassVar,
+    DefaultDict,
+    Dict,
+    Generic,
+    Iterable,
+    List,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+)
 
 import numpy as np
 import pandas as pd
@@ -82,11 +97,16 @@ def _replace_defaults_wrapper(old_init: Callable) -> Callable:
     def new_init(self: BaseTpcpObject, *args: Any, **kwargs: Any) -> None:
         # call the old init.
         old_init(self, *args, **kwargs)
+
         # Check if any of the initial values has a "default parameter flag".
         # If yes we replace it with a clone (in case of a tpcp object) or a deepcopy in case of other objects.
-        for k, v in self.get_params(deep=False).items():
-            if isinstance(v, BaseFactory):
-                setattr(self, k, v.get_value())
+        # This is handled by the factory `get_value` method.
+        for p in get_param_names(self):
+            if isinstance(val := getattr(self, p), BaseFactory):
+                setattr(self, p, val.get_value())
+
+    # This is just for introspection, in case we want to know if we have a modified init.
+    new_init.__tpcp_wrapped__ = True
 
     return new_init
 
@@ -109,8 +129,6 @@ def _retry_eval_with_missing_locals(
             missing = str(e).split("'")[1]
             localns[missing] = None
         except AttributeError as e:
-            if "'NoneType' object has no attribute 'GridSearch'" not in str(e):
-                raise e
             raise RuntimeError(
                 "You ran into an edge case of the builtin type resolver. "
                 "This happens if you use a nested type annotation that is only valid during runtime. "
@@ -152,12 +170,17 @@ def _retry_eval_with_missing_locals(
 
 
 def _custom_get_type_hints(cls: Type[_BaseTpcpObject]) -> Dict[str, Any]:
-    """Extract type hints while avoiding issues with forward references."""
+    """Extract type hints while avoiding issues with forward references.
+
+    We automatically skip all douple_underscore methods.
+    """
     hints = {}
     for base in reversed(cls.__mro__):
         base_globals = sys.modules[base.__module__].__dict__
         ann = base.__dict__.get("__annotations__", {})
         for name, value in ann.items():
+            if name.startswith("__"):
+                continue
             if value is None:
                 value = type(None)
             elif isinstance(value, str):
@@ -174,7 +197,12 @@ def _extract_annotations(
     cls_annotations = _custom_get_type_hints(cls)
     para_annotations = {}
     for k, v in cls_annotations.items():
-        if get_origin(v) is Annotated:
+        origin = get_origin(v)
+        if origin is ClassVar:
+            # If the parameter is a ClassVar, we go one level deeper and check, if its argument was annotated.
+            v = get_args(v)[0]
+            origin = get_origin(v)
+        if origin is Annotated:
             for annot in get_args(v)[1:]:
                 if isinstance(annot, _ParaTypes):
                     para_annotations[k] = annot
@@ -184,31 +212,90 @@ def _extract_annotations(
     return para_annotations
 
 
+def _validate_parameter(instance: _BaseTpcpObject):
+    # We extract the fields of the init
+    fields = _get_init_defaults(type(instance))
+
+    # Validation
+    _has_dangerous_mutable_default(fields, instance)
+    _has_invalid_name(fields, instance)
+
+    if (
+        dataclasses.is_dataclass(instance)
+        or getattr(
+            instance, "__attrs_attrs__", False
+        )  # This checks if the class it an `attrs` class, without importing `attrs`
+    ) and any(isinstance(field.default, BaseFactory) for field in fields.values()):
+        # When you are already using dataclasses or attrs, you must use their factory methods, as we don't overwrite
+        # the init to support CloneFactory anymore.
+        raise ValidationError(
+            "You are using the tpcp default factory (`cf`/`CloneFactory`) in combination with "
+            "`dataclasses` or `attrs`. "
+            "Use the `default_factory`/`factory` option of `dataclasses.field`/`attrs.field` instead."
+        )
+
+
+def _get_tpcp_validated(cls_or_instance: Union[Type[_BaseTpcpObject], _BaseTpcpObject]):
+    cls = cls_or_instance if isinstance(cls_or_instance, type) else type(cls_or_instance)
+    return cls.__dict__.get("__tpcp_validated_hidden__", False)
+
+
+def _set_tpcp_validated(cls_or_instance: Union[Type[_BaseTpcpObject], _BaseTpcpObject], value: bool):
+    cls = cls_or_instance if isinstance(cls_or_instance, type) else type(cls_or_instance)
+    cls.__tpcp_validated_hidden__ = value
+
+
 class _BaseTpcpObject:
-    __field_annotations__: Dict[str, _ParaTypes]
+    # These two parameters should be initialized once per class.
+    # This means, when we read them, we always check in the dictionary if they are defined.
+    __field_annotations_cache__: Dict[str, _ParaTypes]
+    __tpcp_validated_hidden__: bool
+
+    @property
+    def __field_annotations__(self) -> Dict[str, _ParaTypes]:
+        """Get the field annotations that provide higher level information about the role of a parameter.
+
+        We implement that as property to move the extraction of the annotations as far down in the initialization as
+        possible, when we are sure the cls definition is properly finalized (i.e. dataclass decorators are run)
+        """
+        cls = type(self)
+        # We check if we have a cache, and if it is our cache or the cache inherited by the parent class.
+        # If it is from the parent class, we make sure, we create our own.
+        if cache := cls.__dict__.get("__field_annotations_cache__", None):
+            return cache
+        # We need to create a new cache!
+        fields = _get_init_defaults(cls)
+        field_annotations = _extract_annotations(cls, fields)
+        # We only validate the annotations here. If the user is not using the annotations, no need to bother them
+        # with useless error messages
+        _annotations_are_valid(field_annotations, fields, cls_name=cls.__name__)
+        # We set the cache on the cls, so that only one instance is required to process it.
+        cls.__field_annotations_cache__ = field_annotations
+        return field_annotations
 
     def __init_subclass__(cls, *, _skip_validation: bool = False, **kwargs: Any):
         super().__init_subclass__(**kwargs)
-        fields = _get_init_defaults(cls)
+        # Todo: Deprecate this parameter
+        if _skip_validation:
+            warnings.warn(
+                "`_skip_validation` is deprecated and has no effect."
+                "Validation is now performed when the class is actually used and hence has no runtime cost "
+                "during class definition."
+            )
+        # If the class has no init or did inherit its init from the parent (i.e. it is not in its own dict), there is
+        # nothing to do.
+        if "__init__" not in cls.__dict__:
+            return
 
-        cls.__field_annotations__ = _extract_annotations(cls, init_fields=fields)
-
-        # Validation
-        if _skip_validation is not True:
-            _has_dangerous_mutable_default(fields, cls)
-            _has_invalid_name(fields, cls)
-            _annotations_are_valid(fields, cls)
-
-        if cls.__init__ is not object.__init__ and any(
-            isinstance(field.default, BaseFactory) for field in fields.values()
-        ):
-            # In case we have fields that use a factory, we need to wrap the init to replace it
-            setattr(cls, "__init__", _replace_defaults_wrapper(cls.__init__))
+        # If we have a custom init, we need to wrap it with a method that replaces our default values on init.
+        setattr(cls, "__init__", _replace_defaults_wrapper(cls.__init__))
 
 
-class BaseTpcpObject(_BaseTpcpObject, _skip_validation=True):
+class BaseTpcpObject(_BaseTpcpObject):
     """Baseclass for all tpcp objects."""
 
+    _composite_params: ClassVar[Tuple[str, ...]] = tuple()
+
     def get_params(self, deep: bool = True) -> Dict[str, Any]:
         """Get parameters for this algorithm.
 
@@ -227,7 +314,7 @@ class BaseTpcpObject(_BaseTpcpObject, _skip_validation=True):
         """
         return _get_params(self, deep)
 
-    def set_params(self: Self, **params: Any) -> Self:
+    def set_params(self, **params: Any) -> Self:
         """Set the parameters of this Algorithm.
 
         To set parameters of nested objects use `nested_object_name__para_name=`.
@@ -241,6 +328,27 @@ class BaseTpcpObject(_BaseTpcpObject, _skip_validation=True):
         """
         return clone(self, safe=True)
 
+    def __repr_parameter__(self, name: str, value: Any) -> str:
+        """Format a parameter for the __repr__ method.
+
+        This method is used to format the parameters for the __repr__ method.
+        It is called for each parameter and can be overwritten to customize the output.
+
+        Parameters
+        ----------
+        name
+            The name of the parameter
+        value
+            The value of the parameter
+
+        Returns
+        -------
+        str
+            The formatted string for the parameter
+
+        """
+        return f"{name}={value!r}"
+
     def __repr__(self):
         """Provide generic representation for the object based on all parameters."""
         class_name = type(self).__name__
@@ -252,25 +360,125 @@ class BaseTpcpObject(_BaseTpcpObject, _skip_validation=True):
                 first = False
             else:
                 result.append(", ")
-            result.extend((name, "=", repr(para)))
+            result.append(self.__repr_parameter__(name, para))
         return "".join(result) + ")"
 
+    @classmethod
+    def __clone_param__(cls, param_name: str, value: Any) -> Any:  # noqa: unused-argument
+        """Handle cloning of specific parameters of the object.
+
+        This method exists only to allow object the handling of "strange" python datatypes that can not be deepcopied
+        for some reason.
+        This can happen when library authors overwrite deepcopy behaviour or if handling objects can not be easily
+        deepcopied as there are implemented in Cython (or similar).
+
+        In such cases you can overwrite how certain objects should be cloned using this method.
+        Note, that you are on your own, when you do that.
+        I.e. we will not check that you do it correctly and that you ensure that the new object is really independent of
+        the old or any other things you might run into.
+
+        If you are using this method, we recommend to at least check that the cloned object is identical to the
+        original based on our implemented hashing method.
+
+        >>> from tpcp._hash import custom_hash
+        >>> assert custom_hash(my_obj) == custom_hash(my_obj.clone())
+
+        If this doesn't pass, many other features in tpcp will not work as well.
+        """
+        return clone(value, safe=False)
+
+
+def _get_deep_params(obj, parent_key) -> Dict[str, Any]:
+    # This is a little magic, that also gets the parameters of nested sklearn classifiers.
+    if hasattr(obj, "get_params"):
+        deep_items = obj.get_params(deep=True).items()
+        return {parent_key + "__" + k: val for k, val in deep_items}
+    return {}
+
+
+def _assert_is_allowed_composite_value(val, parent_key: str, iteration: int):
+    if (not isinstance(val, tuple)) or (len(val) != 2) or (not isinstance(val[0], str)):
+        raise ValidationError(
+            f"The provided parameters for the composite field {parent_key} does not seem to be the "
+            "right type. "
+            "It should be a sequence of `(name, value)` tuples, but the obj at position "
+            f"{iteration} in the sequence was not a tuple but:\n`{val}`"
+        )
+
 
-# TODO: Make public api
 def _get_params(instance: _BaseTpcpObject, deep: bool = True) -> Dict[str, Any]:
-    valid_fields = get_param_names(type(instance))
+    # At some point, we want to validate that the user defined the class correctly.
+    # To allow for all strange modifications of the class, we run the validation, when we actully need the parameters.
+    # This usually required a call to `get_params`.
+    # Hence, we run the validation here.
+    # To avoid running this validation on every get_params call, we use `__tpcp_validated__` to make sure it runs only
+    # once per class.
+    # Note, that we need to check the class dict and not just the parameter.
+    # Otherwise, the parameter might have been specified on the parent.
+    if not _get_tpcp_validated(instance):
+        _validate_parameter(instance)
+        _set_tpcp_validated(instance, True)
 
+    valid_fields = get_param_names(type(instance))
+    comp_fields = getattr(instance, "_composite_params", tuple())
     out: Dict[str, Any] = {}
     for key in valid_fields:
         value = getattr(instance, key)
-        # This is a little magic, that also gets the parameters of nested sklearn classifiers.
-        if deep and hasattr(value, "get_params"):
-            deep_items = value.get_params(deep=True).items()
-            out.update((key + "__" + k, val) for k, val in deep_items)
+        if deep:
+            if value is not None and comp_fields and key in comp_fields:
+                # Here we handle composite field (i.e. fields that contain sequences of (name, tpcp_obj_instance)
+                # tuples.
+                # We basically flatten the list by storing the instance under the name `{key}__{name}` and then add
+                # all the nested parameters
+                for i, v in enumerate(value):
+                    _assert_is_allowed_composite_value(v, key, i)
+                    name, obj = v
+                    nested_key = f"{key}__{name}"
+                    out[nested_key] = obj
+                    if deep:
+                        out.update(**_get_deep_params(obj, nested_key))
+            else:
+                out.update(**_get_deep_params(value, key))
         out[key] = value
     return out
 
 
+def _set_comp_field(instance, field_name, params):
+    # We first partition our field names to know to which index they belong
+    comp_params: DefaultDict[str, Any] = defaultdict(dict)
+    for key, value in params.items():
+        key, delim, sub_key = key.partition("__")
+        if delim:
+            comp_params[key][sub_key] = value
+        else:
+            comp_params[key]["*"] = value
+
+    # Then we iterate over existing values in the compound field and recreate it
+    new_list = []
+    comp_list = list(getattr(instance, field_name))
+    for key, old_value in comp_list:
+        nested_values = comp_params.pop(key, {})
+        if "*" not in nested_values:
+            new_value = old_value
+        else:
+            new_value = nested_values.pop("*")
+        if nested_values:
+            new_value = new_value.set_params(**nested_values)
+        new_list.append((key, new_value))
+    if comp_params:
+        # Some values are left over -> aka they are invalid
+        raise ValueError(
+            f"You are trying to set values for a compound field {field_name} with the identifiers "
+            f"{list(comp_params.keys())}. "
+            "These identifiers could not be found within the existing list of identifiers for this "
+            f"field ({[c[0] for c in comp_list]}. "
+            "We don't support setting params on not existent value in a compound field. "
+            "Recreate the entire field if you want to add or delete entries from a compound field."
+        )
+
+    setattr(instance, field_name, new_list)
+
+
 def _set_params(instance: BaseTpcpObjectObjT, **params: Any) -> BaseTpcpObjectObjT:
     """Set the parameters of an instance.
 
@@ -280,6 +488,7 @@ def _set_params(instance: BaseTpcpObjectObjT, **params: Any) -> BaseTpcpObjectOb
     if not params:
         return instance
     valid_params = instance.get_params(deep=True)
+    comp_fields = getattr(instance, "_composite_params", tuple())
 
     nested_params: DefaultDict[str, Any] = defaultdict(dict)  # grouped by prefix
     for key, value in params.items():
@@ -294,11 +503,14 @@ def _set_params(instance: BaseTpcpObjectObjT, **params: Any) -> BaseTpcpObjectOb
             valid_params[key] = value
 
     for key, sub_params in nested_params.items():
-        valid_params[key].set_params(**sub_params)
+        if key in comp_fields:
+            _set_comp_field(instance, key, sub_params)
+        else:
+            valid_params[key].set_params(**sub_params)
     return instance
 
 
-def get_param_names(cls: Type[_BaseTpcpObject]) -> List[str]:
+def get_param_names(obj: Union[Type[_BaseTpcpObject], _BaseTpcpObject]) -> List[str]:
     """Get parameter names for the object.
 
     The parameters of an algorithm/pipeline are defined based on its `__init__` method.
@@ -314,6 +526,7 @@ def get_param_names(cls: Type[_BaseTpcpObject]) -> List[str]:
         List of parameter names of the algorithm
 
     """
+    cls = obj if isinstance(obj, type) else type(obj)
     parameters = list(_get_init_defaults(cls).values())
     for p in parameters:
         if p.kind == p.VAR_POSITIONAL:
@@ -326,14 +539,14 @@ def get_param_names(cls: Type[_BaseTpcpObject]) -> List[str]:
 
 
 def _get_annotated_fields_of_type(
-    instance_or_cls: Union[Type[BaseTpcpObject], BaseTpcpObject], field_type: Union[_ParaTypes, Iterable[_ParaTypes]]
+    instance_or_cls: Union[BaseTpcpObject], field_type: Union[_ParaTypes, Iterable[_ParaTypes]]
 ) -> List[str]:
     if isinstance(field_type, _ParaTypes):
         field_type = [field_type]
     return [k for k, v in instance_or_cls.__field_annotations__.items() if v in field_type]
 
 
-def _has_dangerous_mutable_default(fields: Dict[str, inspect.Parameter], cls: Type[_BaseTpcpObject]) -> None:
+def _has_dangerous_mutable_default(fields: Dict[str, inspect.Parameter], instance: _BaseTpcpObject) -> None:
     mutable_defaults = []
 
     for name, field in fields.items():
@@ -344,7 +557,7 @@ def _has_dangerous_mutable_default(fields: Dict[str, inspect.Parameter], cls: Ty
 
     if len(mutable_defaults) > 0:
         raise MutableDefaultsError(
-            f"The class {cls.__name__} contains mutable objects as default values ({mutable_defaults}). "
+            f"The class {type(instance).__name__} contains mutable objects as default values ({mutable_defaults}). "
             "This can lead to unexpected and unpleasant issues! "
             "To solve this the default value should be generated by a factory that produces a new value for each "
             "instance. "
@@ -358,8 +571,10 @@ def _has_dangerous_mutable_default(fields: Dict[str, inspect.Parameter], cls: Ty
         )
 
 
-def _annotations_are_valid(fields: Dict[str, inspect.Parameter], cls: Type[_BaseTpcpObject]) -> None:
-    for k, v in cls.__field_annotations__.items():
+def _annotations_are_valid(
+    field_annotations: Dict[str, _ParaTypes], fields: Dict[str, inspect.Parameter], cls_name: str
+) -> None:
+    for k, v in field_annotations.items():
         if "__" in k:
             if v is _ParaTypes.SIMPLE:
                 warnings.warn(
@@ -369,19 +584,27 @@ def _annotations_are_valid(fields: Dict[str, inspect.Parameter], cls: Type[_Base
                 )
         elif k not in fields:
             raise ValueError(
-                f"The field '{k}' of {cls.__name__} was annotated as a `tpcp` (Hyper/Pure/Normal/Optimizable)-"
-                f"Parameter, but is not a parameter listed in the init! "
+                f"The field '{k}' of {cls_name} was annotated as a `tpcp` "
+                "(Hyper/Pure/Normal/Optimizable)-Parameter, but is not a parameter listed in the init! "
                 "Add the parameter to the init if it is an actual parameter of your algorithm, or remove the "
                 "annotation."
             )
 
 
-def _has_invalid_name(fields: Dict[str, inspect.Parameter], cls: Type[_BaseTpcpObject]) -> None:
+def _has_invalid_name(fields: Dict[str, inspect.Parameter], instance: _BaseTpcpObject) -> None:
     invalid_names = [f for f in fields if "__" in f]
     if len(invalid_names) > 0:
         raise ValidationError(
-            f"The parameters {invalid_names} of {cls.__name__} have a double-underscore in their name. "
-            "This is not allowed, as it interferes with he nested naming conventions in tpcp."
+            f"The parameters {invalid_names} of {type(instance).__name__} have a double-underscore in their name. "
+            "This is not allowed, as it interferes with the nested naming conventions in tpcp.\n\n"
+            "If you are seeing this while using `dataclasses`, and trying to annotate nested parameters, make sure to "
+            "exclude from the init using the explicit `field` syntax or by wrapping it with `ClassVar`:\n\n"
+            ">>> @dataclasses.dataclass\n"
+            "... class MyClass(BaseTpcpObject):\n"
+            "...    # first option\n"
+            "...    nested__parameter: OptiPara[int] = dataclasses.field(init=False)\n"
+            "...    # second option\n"
+            "...    other_nested__parameter: ClassVar[OptiPara[int]]\n\n"
         )
 
 
@@ -436,9 +659,9 @@ def clone(algorithm: T, *, safe: bool = False) -> T:
     # XXX: not handling dictionaries
     if isinstance(algorithm, (list, tuple, set, frozenset)):
         return type(algorithm)([clone(a, safe=safe) for a in algorithm])  # noqa: to-many-function-args
-    # Compared to sklearn, we check specifically for _BaseSerializable and not just if `get_params` is defined on the
+    # Compared to sklearn, we check specifically for BaseTpcpObject and not just if `get_params` is defined on the
     # object.
-    # Due to the way algorithms/pipelines in tpcp work, they need to inherit from _BaseSerializable.
+    # Due to the way algorithms/pipelines in tpcp work, they need to inherit from BaseTpcpObject.
     # Therefore, we check explicitly for that, as we do not want to accidentally treat a sklearn algo (or similar) as
     # algorithm
     if not isinstance(algorithm, BaseTpcpObject):
@@ -453,7 +676,9 @@ def clone(algorithm: T, *, safe: bool = False) -> T:
     klass = algorithm.__class__
     new_object_params = algorithm.get_params(deep=False)
     for name, param in new_object_params.items():
-        new_object_params[name] = clone(param, safe=False)
+        # We defer the actual cloning of the parameters to the class itself to allow to handle stupid edge cases on
+        # user side without modifying tpcp itself.
+        new_object_params[name] = klass.__clone_param__(name, param)
     new_object = klass(**new_object_params)
     params_set = new_object.get_params(deep=False)
 
@@ -478,6 +703,9 @@ class BaseFactory:
         """
         raise NotImplementedError()
 
+    def __call__(self) -> Any:
+        return self.get_value()
+
 
 class CloneFactory(BaseFactory, Generic[T]):
     """Init factory that creates a clone of the provided default values on instance initialisation.
@@ -494,6 +722,10 @@ class CloneFactory(BaseFactory, Generic[T]):
         """Clone the default value for each instance."""
         return clone(self.default_value)
 
+    def __repr__(self) -> str:
+        """Print the representation for the factory."""
+        return f"cf({self.default_value!r})"
+
 
 # The typing here is obviously wrong and a hack to make object that are wrapped by clone factory seem to be just of
 # the same type.
