@@ -1,4 +1,5 @@
 """Base class for all datasets."""
+import dataclasses
 from typing import Dict, Iterator, List, Optional, Sequence, Tuple, TypeVar, Union, cast, overload
 
 import numpy as np
@@ -7,200 +8,13 @@ from typing_extensions import Self
 
 from tpcp._base import BaseTpcpObject
 
-DatasetT = TypeVar("DatasetT", bound="Dataset")
+DatasetT = TypeVar("DatasetT", bound="_Dataset")
 
 
-class Dataset(BaseTpcpObject, _skip_validation=True):
-    """Baseclass for tpcp Dataset objects.
-
-    This class provides fundamental functionality like iteration, getting subsets, and compatibility with `sklearn`'s
-    cross validation helpers.
-
-    For more information check out the examples and user guides on datasets.
-
-    Parameters
-    ----------
-    groupby_cols
-        A column name or a list of column names that should be used to group the index before iterating over it.
-        For examples see below.
-    subset_index
-        For all classes that inherit from this class, subset_index **must** be None by default.
-        But the subclasses require a `create_index` method that returns a DataFrame representing the index.
-
-    Attributes
-    ----------
-    index
-        The index of the dataset.
-        This returns either the `subset_index` or the base index returned by `create_index`.
-    grouped_index
-        The index, but all groupby columns are represented as MultiIndex.
-        Note that the order can be different as the order of index.
-    groups
-        Returns all possible combinations based on the specified `groupby` columns.
-        If `groupby` is None, this returns the row indices.
-        These are also the groups/indices used when iterating over the dataset.
-        The groups are sorted by name.
-    shape
-        Represents the number of all groups encapsulated in a tuple.
-        This is only necessary if `sklearn.model_selection.KFold` is used for splitting the dataset.
-
-    Examples
-    --------
-    This class is usually not meant to be used directly, but the following code snippets show some common operations
-    that can be expected to work for all dataset subclasses.
-
-    >>> import pandas as pd
-    >>> from itertools import product
-    >>>
-    >>> from tpcp import Dataset
-    >>>
-    >>> test_index = pd.DataFrame(
-    ...     list(product(("patient_1", "patient_2", "patient_3"), ("test_1", "test_2"), ("1", "2"))),
-    ...     columns=["patient", "test", "extra"],
-    ... )
-    >>> # We create a little dummy dataset by passing an index directly to `test_index`
-    >>> # Usually we would create a subclass with a `create_index` method that returns a DataFrame representing the
-    >>> # index.
-    >>> dataset = Dataset(subset_index=test_index)
-    >>> dataset
-    Dataset [12 groups/rows]
-    <BLANKLINE>
-             patient    test extra
-       0   patient_1  test_1     1
-       1   patient_1  test_1     2
-       2   patient_1  test_2     1
-       3   patient_1  test_2     2
-       4   patient_2  test_1     1
-       5   patient_2  test_1     2
-       6   patient_2  test_2     1
-       7   patient_2  test_2     2
-       8   patient_3  test_1     1
-       9   patient_3  test_1     2
-       10  patient_3  test_2     1
-       11  patient_3  test_2     2
-
-    We can loop over the dataset.
-    By default, we will loop over each row.
-
-    >>> for r in dataset[:2]:
-    ...     print(r)
-    Dataset [1 groups/rows]
-    <BLANKLINE>
-            patient    test extra
-       0  patient_1  test_1     1
-    Dataset [1 groups/rows]
-    <BLANKLINE>
-            patient    test extra
-       0  patient_1  test_1     2
-
-    We can also change `groupby` (either in the init or afterwards), to loop over other combinations.
-    If we select the level `test`, we will loop over all `patient`-`test` combinations.
-
-    >>> grouped_dataset = dataset.groupby(["patient", "test"])
-    >>> grouped_dataset  # doctest: +NORMALIZE_WHITESPACE
-    Dataset [6 groups/rows]
-    <BLANKLINE>
-                           patient    test extra
-       patient   test
-       patient_1 test_1  patient_1  test_1     1
-                 test_1  patient_1  test_1     2
-                 test_2  patient_1  test_2     1
-                 test_2  patient_1  test_2     2
-       patient_2 test_1  patient_2  test_1     1
-                 test_1  patient_2  test_1     2
-                 test_2  patient_2  test_2     1
-                 test_2  patient_2  test_2     2
-       patient_3 test_1  patient_3  test_1     1
-                 test_1  patient_3  test_1     2
-                 test_2  patient_3  test_2     1
-                 test_2  patient_3  test_2     2
-
-    >>> for r in grouped_dataset[:2]:
-    ...     print(r)  # doctest: +NORMALIZE_WHITESPACE
-    Dataset [1 groups/rows]
-    <BLANKLINE>
-                           patient    test extra
-       patient   test
-       patient_1 test_1  patient_1  test_1     1
-                 test_1  patient_1  test_1     2
-    Dataset [1 groups/rows]
-    <BLANKLINE>
-                           patient    test extra
-       patient   test
-       patient_1 test_2  patient_1  test_2     1
-                 test_2  patient_1  test_2     2
-
-    To iterate over the unique values of a specific level use the "iter_level" function:
-
-    >>> for r in list(grouped_dataset.iter_level("patient"))[:2]:
-    ...     print(r)  # doctest: +NORMALIZE_WHITESPACE
-    Dataset [2 groups/rows]
-    <BLANKLINE>
-                           patient    test extra
-       patient   test
-       patient_1 test_1  patient_1  test_1     1
-                 test_1  patient_1  test_1     2
-                 test_2  patient_1  test_2     1
-                 test_2  patient_1  test_2     2
-    Dataset [2 groups/rows]
-    <BLANKLINE>
-                           patient    test extra
-       patient   test
-       patient_2 test_1  patient_2  test_1     1
-                 test_1  patient_2  test_1     2
-                 test_2  patient_2  test_2     1
-                 test_2  patient_2  test_2     2
-
-    We can also get arbitary subsets from the dataset:
-
-    >>> subset = grouped_dataset.get_subset(patient=["patient_1", "patient_2"], extra="2")
-    >>> subset  # doctest: +NORMALIZE_WHITESPACE
-    Dataset [4 groups/rows]
-    <BLANKLINE>
-                           patient    test extra
-       patient   test
-       patient_1 test_1  patient_1  test_1     2
-                 test_2  patient_1  test_2     2
-       patient_2 test_1  patient_2  test_1     2
-                 test_2  patient_2  test_2     2
-
-    If we want to use datasets in combination with :class:`~sklearn.model_selection.GroupKFold`, we can generate
-    valid group labels as follows.
-
-    .. note::
-        You usually don't want to use that in combination with `self.groupby`.
-
-    >>> # We are using the ungrouped dataset again!
-    >>> group_labels = dataset.create_group_labels(["patient", "test"])
-    >>> pd.concat([dataset.index, pd.Series(group_labels, name="group_labels")], axis=1)
-          patient    test extra         group_labels
-    0   patient_1  test_1     1  (patient_1, test_1)
-    1   patient_1  test_1     2  (patient_1, test_1)
-    2   patient_1  test_2     1  (patient_1, test_2)
-    3   patient_1  test_2     2  (patient_1, test_2)
-    4   patient_2  test_1     1  (patient_2, test_1)
-    5   patient_2  test_1     2  (patient_2, test_1)
-    6   patient_2  test_2     1  (patient_2, test_2)
-    7   patient_2  test_2     2  (patient_2, test_2)
-    8   patient_3  test_1     1  (patient_3, test_1)
-    9   patient_3  test_1     2  (patient_3, test_1)
-    10  patient_3  test_2     1  (patient_3, test_2)
-    11  patient_3  test_2     2  (patient_3, test_2)
-
-    """
-
+class _Dataset(BaseTpcpObject):
     groupby_cols: Optional[Union[List[str], str]]
     subset_index: Optional[pd.DataFrame]
 
-    def __init__(  # noqa: super-init-not-called
-        self,
-        *,
-        groupby_cols: Optional[Union[List[str], str]] = None,
-        subset_index: Optional[pd.DataFrame] = None,
-    ):
-        self.groupby_cols = groupby_cols
-        self.subset_index = subset_index
-
     @property
     def index(self) -> pd.DataFrame:
         """Get index."""
@@ -392,14 +206,18 @@ class Dataset(BaseTpcpObject, _skip_validation=True):
             repr_index = self.index
         else:
             repr_index = self.grouped_index
-        return (
-            f'<h4 style="margin-bottom: 0.1em;">{self.__class__.__name__} [{self.shape[0]} groups/rows]</h3>\n'
-            + repr_index._repr_html_()
+
+        df_repr = (
+            repr_index._repr_html_()
             .replace("<div>", '<div style="margin-top: 0em">')
             .replace('<table border="1" class="dataframe"', '<table style="margin-left: 3em;"')
             .replace("<th>", '<th style="text-align: center;">')
             .replace("<td>", '<td style="text-align: center; padding-left: 2em; padding-right: 2em;">')
         )
+        return (
+            f'<h4 style="margin-bottom: 0.1em;">{self.__class__.__name__} [{self.shape[0]} groups/rows]</h3>\n'
+            + df_repr
+        )
 
     def __iter__(self) -> Iterator[Self]:
         """Return generator object containing a subset for every combination up to and including the selected level."""
@@ -537,6 +355,225 @@ class Dataset(BaseTpcpObject, _skip_validation=True):
         raise NotImplementedError()
 
 
+class Dataset(_Dataset):
+    """Baseclass for tpcp Dataset objects.
+
+    This class provides fundamental functionality like iteration, getting subsets, and compatibility with `sklearn`'s
+    cross validation helpers.
+
+    For more information check out the examples and user guides on datasets.
+
+    Parameters
+    ----------
+    groupby_cols
+        A column name or a list of column names that should be used to group the index before iterating over it.
+        For examples see below.
+    subset_index
+        For all classes that inherit from this class, subset_index **must** be None by default.
+        But the subclasses require a `create_index` method that returns a DataFrame representing the index.
+
+    Attributes
+    ----------
+    index
+        The index of the dataset.
+        This returns either the `subset_index` or the base index returned by `create_index`.
+    grouped_index
+        The index, but all groupby columns are represented as MultiIndex.
+        Note that the order can be different as the order of index.
+    groups
+        Returns all possible combinations based on the specified `groupby` columns.
+        If `groupby` is None, this returns the row indices.
+        These are also the groups/indices used when iterating over the dataset.
+        The groups are sorted by name.
+    shape
+        Represents the number of all groups encapsulated in a tuple.
+        This is only necessary if `sklearn.model_selection.KFold` is used for splitting the dataset.
+
+    Examples
+    --------
+    This class is usually not meant to be used directly, but the following code snippets show some common operations
+    that can be expected to work for all dataset subclasses.
+
+    >>> import pandas as pd
+    >>> from itertools import product
+    >>>
+    >>> from tpcp import Dataset
+    >>>
+    >>> test_index = pd.DataFrame(
+    ...     list(product(("patient_1", "patient_2", "patient_3"), ("test_1", "test_2"), ("1", "2"))),
+    ...     columns=["patient", "test", "extra"],
+    ... )
+    >>> # We create a little dummy dataset by passing an index directly to `test_index`
+    >>> # Usually we would create a subclass with a `create_index` method that returns a DataFrame representing the
+    >>> # index.
+    >>> dataset = Dataset(subset_index=test_index)
+    >>> dataset
+    Dataset [12 groups/rows]
+    <BLANKLINE>
+             patient    test extra
+       0   patient_1  test_1     1
+       1   patient_1  test_1     2
+       2   patient_1  test_2     1
+       3   patient_1  test_2     2
+       4   patient_2  test_1     1
+       5   patient_2  test_1     2
+       6   patient_2  test_2     1
+       7   patient_2  test_2     2
+       8   patient_3  test_1     1
+       9   patient_3  test_1     2
+       10  patient_3  test_2     1
+       11  patient_3  test_2     2
+
+    We can loop over the dataset.
+    By default, we will loop over each row.
+
+    >>> for r in dataset[:2]:
+    ...     print(r)
+    Dataset [1 groups/rows]
+    <BLANKLINE>
+            patient    test extra
+       0  patient_1  test_1     1
+    Dataset [1 groups/rows]
+    <BLANKLINE>
+            patient    test extra
+       0  patient_1  test_1     2
+
+    We can also change `groupby` (either in the init or afterwards), to loop over other combinations.
+    If we select the level `test`, we will loop over all `patient`-`test` combinations.
+
+    >>> grouped_dataset = dataset.groupby(["patient", "test"])
+    >>> grouped_dataset  # doctest: +NORMALIZE_WHITESPACE
+    Dataset [6 groups/rows]
+    <BLANKLINE>
+                           patient    test extra
+       patient   test
+       patient_1 test_1  patient_1  test_1     1
+                 test_1  patient_1  test_1     2
+                 test_2  patient_1  test_2     1
+                 test_2  patient_1  test_2     2
+       patient_2 test_1  patient_2  test_1     1
+                 test_1  patient_2  test_1     2
+                 test_2  patient_2  test_2     1
+                 test_2  patient_2  test_2     2
+       patient_3 test_1  patient_3  test_1     1
+                 test_1  patient_3  test_1     2
+                 test_2  patient_3  test_2     1
+                 test_2  patient_3  test_2     2
+
+    >>> for r in grouped_dataset[:2]:
+    ...     print(r)  # doctest: +NORMALIZE_WHITESPACE
+    Dataset [1 groups/rows]
+    <BLANKLINE>
+                           patient    test extra
+       patient   test
+       patient_1 test_1  patient_1  test_1     1
+                 test_1  patient_1  test_1     2
+    Dataset [1 groups/rows]
+    <BLANKLINE>
+                           patient    test extra
+       patient   test
+       patient_1 test_2  patient_1  test_2     1
+                 test_2  patient_1  test_2     2
+
+    To iterate over the unique values of a specific level use the "iter_level" function:
+
+    >>> for r in list(grouped_dataset.iter_level("patient"))[:2]:
+    ...     print(r)  # doctest: +NORMALIZE_WHITESPACE
+    Dataset [2 groups/rows]
+    <BLANKLINE>
+                           patient    test extra
+       patient   test
+       patient_1 test_1  patient_1  test_1     1
+                 test_1  patient_1  test_1     2
+                 test_2  patient_1  test_2     1
+                 test_2  patient_1  test_2     2
+    Dataset [2 groups/rows]
+    <BLANKLINE>
+                           patient    test extra
+       patient   test
+       patient_2 test_1  patient_2  test_1     1
+                 test_1  patient_2  test_1     2
+                 test_2  patient_2  test_2     1
+                 test_2  patient_2  test_2     2
+
+    We can also get arbitary subsets from the dataset:
+
+    >>> subset = grouped_dataset.get_subset(patient=["patient_1", "patient_2"], extra="2")
+    >>> subset  # doctest: +NORMALIZE_WHITESPACE
+    Dataset [4 groups/rows]
+    <BLANKLINE>
+                           patient    test extra
+       patient   test
+       patient_1 test_1  patient_1  test_1     2
+                 test_2  patient_1  test_2     2
+       patient_2 test_1  patient_2  test_1     2
+                 test_2  patient_2  test_2     2
+
+    If we want to use datasets in combination with :class:`~sklearn.model_selection.GroupKFold`, we can generate
+    valid group labels as follows.
+
+    .. note::
+        You usually don't want to use that in combination with `self.groupby`.
+
+    >>> # We are using the ungrouped dataset again!
+    >>> group_labels = dataset.create_group_labels(["patient", "test"])
+    >>> pd.concat([dataset.index, pd.Series(group_labels, name="group_labels")], axis=1)
+          patient    test extra         group_labels
+    0   patient_1  test_1     1  (patient_1, test_1)
+    1   patient_1  test_1     2  (patient_1, test_1)
+    2   patient_1  test_2     1  (patient_1, test_2)
+    3   patient_1  test_2     2  (patient_1, test_2)
+    4   patient_2  test_1     1  (patient_2, test_1)
+    5   patient_2  test_1     2  (patient_2, test_1)
+    6   patient_2  test_2     1  (patient_2, test_2)
+    7   patient_2  test_2     2  (patient_2, test_2)
+    8   patient_3  test_1     1  (patient_3, test_1)
+    9   patient_3  test_1     2  (patient_3, test_1)
+    10  patient_3  test_2     1  (patient_3, test_2)
+    11  patient_3  test_2     2  (patient_3, test_2)
+
+    """
+
+    def __init__(
+        self,
+        *,
+        groupby_cols: Optional[Union[List[str], str]] = None,
+        subset_index: Optional[pd.DataFrame] = None,
+    ):
+        self.groupby_cols = groupby_cols
+        self.subset_index = subset_index
+
+    @staticmethod
+    def as_dataclass():
+        """Return a version of the Dataset class that can be subclassed using dataclasses."""
+
+        @dataclasses.dataclass(eq=False, repr=False, order=False)
+        class DatasetDc(_Dataset):
+            """Dataclass version of Dataset."""
+
+            groupby_cols: Optional[Union[List[str], str]] = None
+            subset_index: Optional[pd.DataFrame] = None
+
+        return DatasetDc
+
+    @staticmethod
+    def as_attrs():
+        """Return a version of the Dataset class that can be subclassed using `attrs` defined classes.
+
+        Note, this requires `attrs` to be installed!
+        """
+        from attrs import define  # noqa: import-outside-toplevel
+
+        @define(eq=False, repr=False, order=False, kw_only=True, slots=False)
+        class DatasetAt(_Dataset):
+            """Attrs version of Dataset."""
+
+            groupby_cols: Optional[Union[List[str], str]] = None
+            subset_index: Optional[pd.DataFrame] = None
+
+        return DatasetAt
+
+
 T = TypeVar("T")
 
 
